"""
This file contains functions and classes to perform a sourcetime function
inversion using the projected landweber method as shown to work for seismology
by Bertero et al. as well as Valleé et al.

:copyright:
    Lucas Sawade (lsawade@princeton.edu, 2019)

:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)

"""

from numpy.fft import fft, ifft, fftfreq
import matplotlib.pyplot as plt
import numpy as np
from copy import deepcopy as dc


def conj_grad(G, u, t, dt, tau_factor=1, iT=0, fT=120, niter=10):
    """Computing one landweber iteration of the signal.
    param. Refer back to Bertero et al."""

    # Check which is longer
    N = len(G)

    # Positive part of t
    # Fs = 1 / (N * dt)
    tshift = -np.min(t)
    freq = fftfreq(N, d=dt)
    eq_shift = np.exp(-1.j * freq * 2 * np.pi * tshift)

    # Fourier Transform stuff
    Gw = fft(G)
    uw = fft(u)

    # Create first iteration
    # Create first iteration
    # f = correlate(u, G, 'same')
    f = np.real(ifft(uw/Gw*eq_shift))

    plt.figure()
    plt.plot(f)

    fw = fft(f)

    # Empty list for all iterations
    ft = []
    fwt = []

    # Actual Method
    r = uw - Gw * fw
    p = r
    rsold = np.dot(r, np.conjugate(r))

    for i in range(niter):
        ffp = fft(p)
        Ap = Gw * ffp
        alpha = rsold / (np.conjugate(ffp) * Ap)

        g = np.real(ifft((f + alpha * p) * eq_shift))

        f = lw_projection(g, t, iT=iT, fT=fT)
        ft.append(f)
        fw = fft(f)
        fwt.append(fw)

        r = r - alpha * Ap

        rsnew = np.dot(r, np.conjugate(r))

        if np.sqrt(rsnew) < 1e-10:
            break

        p = r + (rsnew / rsold) * p

        rsold = rsnew

    return fwt, ft


def landweber(Gw, uw, tau, f, eq_shift):
    """Computing one landweber iteration of the signal.
    param. Refer back to Bertero et al."""

    # We have to shift fw in time to account for the negative part
    # time values in the signal
    gp1 = f + tau * np.real(ifft((np.conj(Gw) * eq_shift * (uw - Gw * fft(
        f)))))

    return gp1


def lw_projection(g, t, iT=2, fT=120):
    """Projects a trace into the causal non-negative Convex space."""

    # Copy g
    f = dc(g)

    # Apply projection parameters
    f[np.where(t < iT)[0]] = 0
    f[np.where(t > fT)[0]] = 0
    f[np.where(f < 0)[0]] = 0

    return f


def projected_landweber(G, u, t, dt, tau_factor=1, iT=0, fT=120, niter=10):

    # Check which is longer
    N = len(G)

    # Positive part of t
    # Fs = 1/(N*dt)
    tshift = -np.min(t)
    freq = fftfreq(N, d=dt)
    eq_shift = np.exp(-1.j*freq*2*np.pi*tshift)

    # Fourier Transform stuff
    Gw = fft(G)
    uw = fft(u)

    # Create first iteration
    fp1 = np.zeros(Gw.size)

    # Compute tau
    tau = 0.05/np.max(np.abs(Gw))

    # Empty list for all iterations
    ft = []
    fwt = []

    for i in range(niter):

        # Compute the landweber iteration
        g = landweber(Gw, uw, tau, fp1, eq_shift)

        # Compute the projection onto the convex set.
        fp1 = lw_projection(g, t, iT=iT, fT=fT)

        # Normalize by integrated sum
        fp1 = fp1
        # Save STF iteration
        ft.append(fp1)

        # Save STF Frequency spectrum
        fwt.append(fft(fp1))

    return fwt, ft


def l2_norm(x):
    return np.sqrt(np.sum(x ** 2))


def compute_error(t, dt, stf_list, green, disp):
    """Compute both euclidean norm for every iteration as
    well as relative error between iterations"""

    N = len(green)

    # Positive part of t
    tshift = np.min(t)

    # Get Frequencies
    freq = fftfreq(N, dt)
    eq_shift = np.exp(-1.j*freq*2*np.pi*tshift)

    euc = np.zeros(len(stf_list))
    relerr = np.zeros(len(stf_list))
    for _i, stf in enumerate(stf_list):

        Ag = np.real(ifft(fft(green) * fft(stf) * eq_shift))
        # Frobenius norm between data and simulation
        euc[_i] = l2_norm(np.abs(np.real(Ag) - disp)) / l2_norm(disp)

        # Frobenius norm between iterations
        if _i > 0:
            relerr[_i] = l2_norm(stf_list[_i] - stf_list[_i - 1]) / l2_norm(
                stf_list[_i - 1])

    relerr[0] = 1

    return euc, relerr


class Operator(object):
    """Linear operator"""

    def __init__(self, t, wavelet):

        # Time Vector
        self.t = t
        self.dt = t[1] - t[0]
        self.N = len(t)

        # Corresponding frequency
        self.freq = fftfreq(self.N, self.dt)

        # possible timeshift due to change in from zero
        if np.min(t) != 0:
            self.tshift = -np.min(t)
            self.eq_shift = np.exp(-1.j*self.freq*2*np.pi*self.tshift)
        else:
            self.eq_shift = 1

        # Time Vector
        self.wavelet = wavelet

    def forward(self, m):
        """Forward operator

        :param m: model
        :return: conveolved data
        """
        return np.real(ifft(fft(self.wavelet)*fft(self.m)*self.eq_shift))

    def adjoint(self, m):
        """Adjoint operator

        :param model: model
        :return: convolved data

        """
        return np.real(ifft(fft(self.wavelet) * np.conjugate(fft(self.m))
                            * self.eq_shift))


class stf_operator(object):
    """Linear operator"""

    def __init__(self, t, g):
        """

        :param t: time vector
        :param g: greens function
        """

        # Time Vector
        self.t = t
        self.dt = t[1] - t[0]
        self.N = len(t)

        # Corresponding frequency
        self.freq = fftfreq(self.N, self.dt)

        # possible timeshift due to change in from zero
        if np.min(t) != 0:
            self.tshift = -np.min(t)
            self.eq_shift = np.exp(-1.j*self.freq*2*np.pi*self.tshift)
        else:
            self.eq_shift = 1

        # Time Vector
        self.g = g

    def forward(self, wavelet):
        """Forward operator

        :param m: model
        :return: conveolved data
        """
        return np.real(ifft(fft(wavelet)*fft(self.g)*self.eq_shift))

    def adjoint(self, wavelet):
        """Adjoint operator

        :param model: model
        :return: convolved data

        """
        return np.real(ifft(fft(wavelet) * np.conjugate(fft(self.g))
                            * self.eq_shift))


class stf_inversion(object):
    """Defines inversion operation."""

    def __init__(self, t, d, g, iT, fT):
        """ Initialize the stf_inversion class

        :param d: data
        :param g: green's function
        """
        self.t = t
        self.dt = t[1] - t[0]
        self.iT = iT
        self.fT = fT
        self.d = d
        self.g = g
        self.n = len(d)
        self.F = stf_operator(t, g)

    def proj_lw(self):
        """Inverts data using projected Landweber method"""

        return projected_landweber(self.g, self.d, self.t, self.dt,
                                   tau_factor=1, iT=0, fT=10, niter=20)

    def conjugate_gradient(self):
        """inverts data"""

        # List to save iterations
        stf_list = []
        fft_list = []
        cost_list = []

        # First Model estimate
        stf = np.zeros_like(self.d)

        # Compute residual (cost function)
        r = self.d - self.F.forward(stf)

        # gradient direction
        s = np.zeros_like(self.d)

        # Initialize beta
        beta = 0

        for i in range(215):

            # Compute gradient
            g = self.F.adjoint(r)

            #
            gamma = np.dot(g, g)

            if i != 0:
                beta = np.dot(g, g) / gamma

            # Compute decent direction
            s = g + beta * s

            # delta (cost function)
            deltar = self.F.forward(s)

            # Scaling factor
            alpha = gamma / np.dot(deltar, deltar)

            # decent
            stf = lw_projection(stf + alpha * s,  self.t,
                                iT=self.iT, fT=self.fT)

            # stf = stf + alpha * s
            fft_list.append(stf)

            # cost
            r = r - alpha * deltar

            stf_list.append(stf)
            cost_list.append(r)

        return fft_list, stf_list, cost_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This script will download the observed data. To the necessary places.


:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""


from gcmt3d.data import DataRequest
from gcmt3d.asdf.utils import smart_read_yaml, is_mpi_env
import os

# Get logger to log progress
from gcmt3d import logger


def data_request(cmt_filename, param_path):

    # Set directories of the parameter files
    request_param_path = os.path.join(param_path,
                                      "RequestParams/RequestParams.yml")

    # Read the parameter file
    rCparams = smart_read_yaml(request_param_path, mpi_mode=is_mpi_env())

    # Earthquake and Station parameters
    cmt_dir = os.path.dirname(cmt_filename)
    station_dir = os.path.join(cmt_dir, "station_data")

    # Get STATIONS file from CMT directory
    stationsfile = os.path.join(station_dir, "STATIONS")

    # Create Request Object
    Request = DataRequest.from_file(cmt_filename,
                                    stationlistfname=stationsfile,
                                    sfstationlist=True,
                                    duration=rCparams['duration'],
                                    channels=rCparams['channels'],
                                    locations=rCparams['locations'],
                                    starttime_offset=rCparams[
                                        'starttime_offset'],
                                    outputdir=cmt_dir)

    # Print Earthquake Download Info
    for line in Request.__str__().splitlines():
        logger.info(line)

    # Request download
    Request.download()



"""

This file contains functions to compute statistics

:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU General Public License, Version 3
    (http://www.gnu.org/copyleft/gpl.html)

Last Update: November 2019

"""

import logging
from glob import glob
import os
import numpy as np
from pycmt3d.source import CMTSource
from ..plot.stats import PlotStats
from ..utils.io import load_json
from ..log_util import modify_logger

logger = logging.getLogger(__name__)
modify_logger(logger)


def read_specfem_station_list(filename):
    """Gets stationlist from file

    Args:
        filename (string): specfem station filename

    Returns:
        list of arguments for each station

    Station list:
        [network station latitude longitude elevation]
    """

    stationlist = []
    with open(filename, 'r') as specfemfile:

        for line in specfemfile:
            # Read stations into list of stations
            line = line.split()
            # Append the [network station latitude longitude elevation]
            newline = [line[1], line[0], float(line[2]), float(line[3]),
                       float(line[4]), float(line[5])]
            # to the station list
            stationlist.append(newline)

    return stationlist


class Struct(object):
    def __init__(self, **entries):
        self.__dict__.update(entries)
        for k, v in self.__dict__.items():
            if isinstance(v, dict):
                self.__dict__[k] = Struct(**v)


def get_stats_json(filename):
    """Reads the stats json and outputs dictionary with
    all the necessary data
    """
    d = load_json(os.path.abspath(filename))
    data_container = {"sta_lon": np.array(d["sta_lon"]),
                      "sta_lat": np.array(d["sta_lat"]),
                      "nwindows": d["nwindows"],
                      "nwin_on_trace": d["nwin_on_trace"]}
    cmtsource = CMTSource.from_dictionary(d["oldcmt"])
    new_cmtsource = CMTSource.from_dictionary(d["newcmt"])
    config = Struct(**d["config"])
    nregions = d["nregions"]
    bootstrap_mean = np.array(d["bootstrap_mean"])
    bootstrap_std = np.array(d["bootstrap_std"])
    var_reduction = d["var_reduction"]
    mode = d["mode"]
    G = d["G"]
    stats = d["stats"]

    return (data_container, cmtsource, new_cmtsource,
            config, nregions, bootstrap_mean,
            bootstrap_std, var_reduction, mode,
            G, stats)


class Statistics(object):
    """Governs the statistics of multiple inversions"""

    def __init__(self, old_cmts, old_ids, new_cmts, new_ids, stations,
                 npar=9, stat_dict: dict or None = None):
        """ Initialize Statistics class

        Args:
        -----
        old_cmts (numpy.ndarray): matrix with old CMT data
        old_ids (list): List of event ids corresponding to matrix rows
        new_cmts (numpy.ndarray): matrix with new CMT data
        new_ids (list): List of event ids corresponding to matrix rows
        stations: list of stations
        npar: number of parameters to perform the analysis on
        verbose: Set verbosity

        Returns:
        --------
        Statistics class containing all necessary data and tools to perform
        an analysis on the data.

        The matrices below should have following columns:
        -------------------------------------------------
        M0, Mrr, Mtt, Mpp, Mrt, Mrp, Mtp, depth, lat, lon, CMT, t_shift, hdur

        """
        self.ocmt = old_cmts
        self.oids = np.array(old_ids)
        self.ncmt = new_cmts
        self.nids = np.array(new_ids)
        self.npar = npar

        # Sanity check
        if any([False for a, b in zip(self.oids, self.nids) if a == b]):
            raise ValueError("Can only compare equal earthquakes.")

        self.ids = self.nids
        self.stations = stations

        # Compute difference/evolution
        self.dCMT = self.ncmt - self.ocmt
        self.dCMT[:, :7] /= self.ocmt[:, 0, None]

        good_stats = []
        for _i, (dcmt, id) in enumerate(zip(self.dCMT, self.ids)):
            if dcmt[0] > 0.5 or dcmt[0] < -0.5 or \
                    dcmt[7] > 20000 or dcmt[7] < -20000:
                logger.info("ID: %s" % id)
                logger.info("  M0_0: %e" % self.ocmt[_i, 0])
                logger.info("  M0_1: %e" % self.ncmt[_i, 0])
                logger.info("  dCMT: %f" % dcmt[0])
                logger.info("  dz: %f" % dcmt[7])
                logger.info("Removed C%s from matrix." % id)
            else:
                good_stats.append(_i)

        # Fix outliers
        good_stats = np.array(good_stats)
        self.ocmt = self.ocmt[good_stats, :]
        self.oids = self.oids[good_stats]
        self.ncmt = self.ncmt[good_stats, :]
        self.nids = self.nids[good_stats]
        self.dCMT = self.dCMT[good_stats, :]

        # Compute Correlation Coefficients
        self.xcorr_mat = np.corrcoef(self.dCMT.T)

        # Compute Mean
        self.mean_mat = np.mean(self.dCMT, axis=0)

        self.mean_dabs = np.mean(np.abs(self.dCMT), axis=0)
        # print(self.mean_dabs)

        # Compute Standard deviation
        self.std_mat = np.std(self.dCMT, axis=0)

        # Create labels
        self.labels = [r"$M_0$",
                       r"$M_{rr}$", r"$M_{tt}$",
                       r"$M_{pp}$", r"$\delta M_{rt}$",
                       r"$M_{rp}$", r"$M_{tp}$",
                       r"$z$", r"Lat", r"Lon",
                       r"$t_{CMT}$", r"$t_{shift}$", r"$hdur$"]

        self.dlabels = [r"$\delta M_0/M_0$",
                        r"$\delta M_{rr}/M_0$", r"$\delta M_{tt}/M_0$",
                        r"$\delta M_{pp}/M_0$", r"$\delta M_{rt}/M_0$",
                        r"$\delta M_{rp}/M_0$", r"$\delta M_{tp}/M_0$",
                        r"$\delta z$", r"$\delta$Lat", r"$\delta$Lon",
                        r"$\delta t_{CMT}$", r"$\delta t_{shift}$",
                        r"$\delta hdur$"]

        self.stat_dict = stat_dict

    @classmethod
    def _from_dir(self, directory, direct=False, npar=9):
        """Load old and new inverted CMTSOLUTIONS into lists of CMTSources.

        Args:
        -----
            database_dir (string): database directory containing all the CMT
                                    solutions

        Returns:
        --------
            tuple of two lists containing the original cmtsolution and its
            corresponding inversion

        """

        logger.info("Looking for earthquakes here: %s" % directory)
        # Get list of inversion files.
        if direct:
            clist = glob(os.path.join(directory, "*.json"))
        else:
            clist = glob(os.path.join(directory, "C*",
                                      "inversion", "inversion_output",
                                      "*.json"))

        logger.info("Found all files.")
        logger.info(" ")

        # Old CMTs
        old_cmts = []
        new_cmts = []
        stat_dicts = []
        station_list = set()

        logger.info("Reading individual files ...")
        logger.info(" ")

        # Loop of files
        for invdata in clist:

            try:
                logger.info("  File: %s" % invdata)

                (data_container,
                 cmtsource,
                 new_cmtsource,
                 config,
                 nregions,
                 bootstrap_mean,
                 bootstrap_std,
                 var_reduction,
                 mode,
                 G,
                 stats) = get_stats_json(invdata)

                # Append CMT files
                old_cmts.append(cmtsource)
                new_cmts.append(new_cmtsource)
                stat_dicts.append(stats)

                # stations
                for lat, lon in zip(data_container["sta_lat"],
                                    data_container["sta_lon"]):
                    station_list.add((lat, lon))

            except Exception as e:
                logging.warning(e)

        logger.info(" ")
        logger.info("Done.")
        logger.info(" ")

        old_cmt_mat, old_ids = Statistics.create_cmt_matrix(old_cmts)
        new_cmt_mat, new_ids = Statistics.create_cmt_matrix(new_cmts)

        complete_dict = self.compute_change(stat_dicts)

        return self(old_cmt_mat, old_ids, new_cmt_mat, new_ids,
                    list(station_list), npar=npar, stat_dict=complete_dict)

    def plot_changes(self, savedir=None):
        """This function plots and saves the plots with the statistics.

        Args:
            savedir (str, optional): Sets directory where to save the
                                     plots. If None the plots will be
                                     displayed. Defaults to None.
        """

        PS = PlotStats(ocmt=self.ocmt, ncmt=self.ncmt, dCMT=self.dCMT,
                       xcorr_mat=self.xcorr_mat, mean_mat=self.mean_mat,
                       std_mat=self.std_mat, stat_dict=self.stat_dict,
                       labels=self.labels, dlabels=self.dlabels,
                       stations=self.stations, npar=self.npar,
                       savedir=savedir)

        PS.plot_main_stats()
        PS.plot_dM_dz()
        PS.plot_dM_z()
        # PS.plot_changes()
        # PS.plot_xcorr_matrix()
        # PS.plot_xcorr_heat()
        # PS.plot_measurement_changes()
        # PS.plot_mean_measurement_change_stats()
        # PS.save_table()

    @staticmethod
    def get_PTI_from_cmts(cmt_mat):
        """ Computes the Principal axes and components
        and gets the Isotropic component

        Args:
            cmt_mat (numpy.ndarray): Matrix of of cmt solution.
                                     Nx6
        Returns: Nx3x3 matrix and
        """
        pass

        """  for cmt in cmt_mat:
        m = np.array([cmt[0]]) """

    @staticmethod
    def compute_change(stat_dicts: list):
        """ Takes in the measurement statistics for and compiles them in
        vectors.

        :param stat_dicts:
        :return: tuple of dicts with before and after data

        """

        complete_dict = dict()

        for _sdict in stat_dicts:
            for tag, measurement_dict in _sdict.items():
                if tag not in complete_dict.keys():
                    complete_dict[tag] = dict()

                for measurement, ba_dict in measurement_dict.items():
                    if measurement not in complete_dict[tag].keys():
                        complete_dict[tag][measurement] = \
                            {"before": {"std": [], "mean": []},
                             "after": {"std": [], "mean": []}}

                    # Add STD
                    for time in ['before', 'after']:
                        complete_dict[tag][measurement][time]["std"].append(
                            ba_dict[time]["std"])
                        complete_dict[tag][measurement][time]["mean"].append(
                            ba_dict[time]["mean"])

        return complete_dict

    @staticmethod
    def create_cmt_matrix(cmt_source_list):
        """Takes in list of CMT sources and transforms it into a
        :class:`numpy.ndarray`

        Args:
        -----
            CMT source list

        Returns:
        --------
            :class:`numpy.ndarray`

        """

        # Number of Earthquakes
        N = len(cmt_source_list)

        # Create empty array for the values
        event_id = []

        # Initialize empty matrix
        cmt_mat = np.zeros((N, 13))

        for _i, cmt in enumerate(cmt_source_list):

            # Populate id list with ids
            event_id.append(cmt.eventname)

            # Populate CMT matrix
            cmt_mat[_i, :] = np.array([cmt.M0, cmt.m_rr, cmt.m_tt, cmt.m_pp,
                                       cmt.m_rt,
                                       cmt.m_rp, cmt.m_tp, cmt.depth_in_m,
                                       cmt.latitude, cmt.longitude,
                                       cmt.cmt_time,
                                       cmt.half_duration,
                                       cmt.time_shift,
                                       ])

        return cmt_mat, event_id


if __name__ == "__main__":

    # Load shit
    database_dir = "/Users/lucassawade/straverse/database"

    ST = Statistics._from_dir(database_dir)
    ST.plot_changes(savedir="/Users/lucassawade")



"""

This script contains functions to create pycmt3d inversion dictionaries. They
have to have following setup:

:param asdf_file_dict: the dictionary which provides the path
                       information of asdf file, for example:
                       {"obsd":"/path/obsd/asdf", "synt":"/path/synt/asdf",
                       "Mrr": "/path/Mrr/synt/asdf", ...}

:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU General Public License, Version 3
    (http://www.gnu.org/copyleft/gpl.html)

Last Update: June 2019

"""

import os
import glob
import sys

from .create_process_paths import get_processing_list
from .create_process_paths import get_windowing_list
from gcmt3d.asdf.utils import write_yaml_file

import logging
from ...log_util import modify_logger

# Create logger
logger = logging.getLogger(__name__)
modify_logger(logger)

PARMAP = {"Mrr": "CMT_rr",
          "Mtt": "CMT_tt",
          "Mpp": "CMT_pp",
          "Mrt": "CMT_rt",
          "Mrp": "CMT_rp",
          "Mtp": "CMT_tp",
          "dep": "CMT_depth",
          "lon": "CMT_lon",
          "lat": "CMT_lat",
          "ctm": "CMT_ctm",
          "hdr": "CMT_hdr"}


def create_inversion_dict(processed_dir, bandstring):
    """ This creates a dictionary with all files necessary simulation files for
    the inversion.

    :param npar: Number of parameters to be inverted.
    :param processed_dir: directory with the processed data.
    :param bandstring: string that is contained in the file name and checks
                       for the bandpass filtering.
    :return: dictionary with data files

    """

    # Get absolute path of process dir in case it wasn't given
    processed_dir = os.path.abspath(processed_dir)

    # Find files in processed seismos directory

    obsd_file = glob.glob(os.path.join(processed_dir,
                                       "*observed." + bandstring + ".h5"))[0]
    synt_file = glob.glob(os.path.join(processed_dir,
                                       "*synthetic_CMT."
                                       + bandstring + ".h5"))[0]

    # Initialize filedict
    filedict = dict()

    # Set observed filename
    filedict["obsd"] = obsd_file

    # Set observed filename
    filedict["synt"] = synt_file

    # Creating a dictionary from synthetic files and their names
    for key, value in PARMAP.items():

        # Get perturbed file with value
        pert_file = glob.glob(os.path.join(processed_dir,
                                           "*synthetic*" + value + "*"
                                           + bandstring + "*.h5"))

        if len(pert_file) != 0:
            # Write into file dict
            filedict[key] = pert_file[0]

    return filedict


def create_full_inversion_dict_list(cmt_file_db,  process_obs_dir,
                                    process_syn_dir, window_process_dir):
    """
    :param cmt_file_db: the cmtsolution file in the database.
    :type cmt_file_db: str
    :param process_obs_dir: directory with the process parameter files for
                            the observed data
    :type cmt_file_db: str
    :param process_syn_dir: directory with the process parameter files for
                            the synthetic data
    :type process_syn_dir: str
    :param window_process_dir: path to the process directory
    :type window_process_dir: str
    :param npar: Number of parameters to invert for
    :type npar: int
    :param verbose: if True verbose output
    :type verbose: bool

    :return: list of full inversion dictionaries


    The inversion dictionaries contain all necessary info to start the
    inversion. Each Inversion dictionary has then a respective band (and
    possible wavetype).

    **Dictionary structure:**

    .. code-block:: yaml

        Window_file: "path/to/window/file"
        ASDF_dict:

            Mrr: "path/to/CMT_rr.h5"
            Mtt: "path/to/CMT_tt.h5"
            Mpp: "path/to/CMT_pp.h5"
            Mrt: "path/to/CMT_rt.h5"
            Mrp: "path/to/CMT_rp.h5"
            Mtp: "path/to/CMT_tp.h5"
            dep: "path/to/CMT_dep.h5"
            lon: "path/to/CMT_lon.h5"
            lat: "path/to/CMT_lat.h5"
            ctm: "path/to/CMT_ctm.h5"
            hdr: "path/to/CMT_hdr.h5"

    """

    # Get CMT dir
    cmt_dir = os.path.dirname(cmt_file_db)

    # inversion dictionary
    output_dir = os.path.join(cmt_dir, "inversion", "inversion_dicts")

    # Get the window file list
    window_path_list, win_list = get_windowing_list(cmt_file_db,
                                                    window_process_dir)

    # Get the processing list
    processing_list, obs_list, syn_list = get_processing_list(cmt_file_db,
                                                              process_obs_dir,
                                                              process_syn_dir)

    # Create empty list of dictionaries
    inv_dict_list = []
    outfile_list = []

    for window_file in win_list:

        # Get Basename
        window_file_name = os.path.basename(window_file)

        # extract info
        band_and_wave_type = (window_file_name.split(".")[1]).split("#")

        # Get Passband
        band = band_and_wave_type[0]

        # Create empty dictionary
        inv_dict = dict()

        # Create Inversion Dictionary
        inv_dict["window_file"] = window_file

        # Add ASDF dictionary
        # Initialize filedict
        asdf_dict = dict()

        # Set observed filename
        asdf_dict["obsd"] = [s for s in obs_list
                             if band in os.path.basename(s)][0]

        # Set observed filename
        asdf_dict["synt"] = [s for s in syn_list
                             if (band in os.path.basename(s))
                             and ("synthetic_CMT." in os.path.basename(s))][0]

        # Creating a dictionary from synthetic files and their names
        for key, value in PARMAP.items():

            # Get perturbed file with value
            pert_file = [s for s in syn_list
                         if (band in os.path.basename(s))
                         and ("synthetic_%s." % value
                              in os.path.basename(s))]

            if len(pert_file) > 1:
                logger.error("Found %d synthetic perturbation files."
                             % len(pert_file))
                raise ValueError

            elif len(pert_file) != 0:
                # Write into file dict
                asdf_dict[key] = pert_file[0]

        inv_dict["asdf_dict"] = asdf_dict

        # Append dictionary to inversion dictionary list.
        inv_dict_list.append(inv_dict)

        output_file_name = os.path.join(output_dir,
                                        "inversion" + window_file_name[7:])

        outfile_list.append(output_file_name)

    return inv_dict_list, outfile_list


def create_g3d_inversion_dict_list(cmt_file_db,  process_obs_dir,
                                   process_syn_dir, window_process_dir):
    """
    :param cmt_file_db: the cmtsolution file in the database.
    :type cmt_file_db: str
    :param process_obs_dir: directory with the process parameter files for
                            the observed data
    :type cmt_file_db: str
    :param process_syn_dir: directory with the process parameter files for
                            the synthetic data
    :type process_syn_dir: str
    :param window_process_dir: path to the process directory
    :type window_process_dir: str
    :param npar: Number of parameters to invert for
    :type npar: int
    :param verbose: if True verbose output
    :type verbose: bool

    :return: list of full inversion dictionaries


    The inversion dictionaries contain all necessary info to start the
    inversion. Each Inversion dictionary has then a respective band (and
    possible wavetype).

    **Dictionary structure:**
    .. code-block:: yaml

        Window_file: "path/to/window/file"
        ASDF_dict:

            Mrr: "path/to/observed.h5"
            Mtt: "path/to/synthetic.h5"


    """

    # Get CMT dir
    cmt_dir = os.path.dirname(cmt_file_db)

    # New synt dir =
    new_synt_dir = os.path.join(cmt_dir, "inversion", "inversion_output",
                                "cmt3d", "new_synt")

    # inversion dictionary
    output_dir = os.path.join(cmt_dir, "inversion", "inversion_dicts")

    # Get the window file list
    window_path_list, win_list = get_windowing_list(cmt_file_db,
                                                    window_process_dir)

    # Get the processing list
    processing_list, obs_list, syn_list = get_processing_list(cmt_file_db,
                                                              process_obs_dir,
                                                              process_syn_dir)

    syn_list = glob.glob(os.path.join(new_synt_dir, "*synt*h5"))

    # Create empty list of dictionaries
    inv_dict_list = []
    outfile_list = []

    for window_file in win_list:

        # Get Basename
        window_file_name = os.path.basename(window_file)

        # extract info
        band_and_wave_type = (window_file_name.split(".")[1]).split("#")

        # Get Passband
        band = band_and_wave_type[0]

        # Create empty dictionary
        inv_dict = dict()

        # Create Inversion Dictionary
        inv_dict["window_file"] = window_file

        # Add ASDF dictionary
        # Initialize filedict
        asdf_dict = dict()

        # Set observed filename
        asdf_dict["obsd"] = [s for s in obs_list
                             if band in os.path.basename(s)][0]

        "200709121110A.9p_ZT.040_100_synt.h5"
        "200709121110A.9p_ZT.090_250_synt.h5"

        # Set observed filename
        asdf_dict["synt"] = [s for s in syn_list
                             if (band in os.path.basename(s))
                             and ("synt." in os.path.basename(s))][0]

        inv_dict["asdf_dict"] = asdf_dict

        # Append dictionary to inversion dictionary list.
        inv_dict_list.append(inv_dict)

        output_file_name = os.path.join(output_dir,
                                        "grid" + window_file_name[7:])

        outfile_list.append(output_file_name)

    return inv_dict_list, outfile_list


def write_inversion_dicts(dict_list, filename_list):
    """ Writes inversion dictionaries to the given output directory

    :param dict_list: List of inversion dictionaries

    :param filename_list: list of paths in string format corresponding to
                          writing location of the dictionaries in the list

    """

    # Loop over lists.
    for inv_dict, filename in zip(dict_list, filename_list):

        # Write file
        write_yaml_file(inv_dict, filename)


if __name__ == "__main__":
    for key, value in create_inversion_dict(sys.argv[1], "017_040").items():
        print(key, value)



"""

This module includes a custom logger to make it easier to identify errors and
debug etc.

The module adds both another level to the logger and the corresponding
formatter. If you want to remove or add any logging level make sure to edit
both the CustomLogger and the CustomFormatter to accommodate your changes!

:author:
    Lucas Sawade (lsawade@princeton.edu, 2019)

:copyright:
    Use it copy it and do whatever you like with it.

"""
import logging

from logging import getLoggerClass
from logging import NOTSET

VERBOSE = 15


def addLoggingLevel(levelName, levelNum, methodName=None):
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example
    -------
    >>> addLoggingLevel('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5


    Taken from StackOverflow because the code was beautifully simple.
    Author: Mad Physicist (Mar 4, 2016)

    """
    if not methodName:
        methodName = levelName.lower()

    if hasattr(logging, levelName):
        return
    if hasattr(logging, methodName):
        return
    if hasattr(logging.getLoggerClass(), methodName):
        return

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def logForLevel(self, message, *args, **kwargs):
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)

    def logToRoot(message, *args, **kwargs):
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)


class CustomLogger(getLoggerClass()):
    """
    This class is just created ot add the VERBOSE level. More
    level could be added to this class to accommodate other levels.

    The variable VERBOSE is given at the top of the module. That way it can be
    changed for all depending function

    The class makes it possible to add extra levels to the classic logger
    The line `addLoggingLevel("VERBOSE", VERBOSE)` in the initalization is
    an example on how to add a level using the `addLoggingLevel` function
    located in this module.

    Don't forget to edit the `CustomFormatter` to accommodate for your
    introduced levels if you are using the `CustomFormatter`.
    An example is given in the class under `# EXTRA LEVELS` for the
    `VERBOSE` level.

    """

    def __init__(self, name, level=NOTSET):
        super().__init__(name, level)

        addLoggingLevel("VERBOSE", VERBOSE)


class CustomFormatter(logging.Formatter):
    """
    Logging Formatter to add colors and count warning / errors

    This class organizes the customization of the logging output.
    The formatter as of now outputs the logs in the following manner in
    order of Loglevel:

    .. code-block:: python

        [2020-03-30 21:00:17] matpy -- [  DEBUG   ] \
                        Test Verbose Level (matrixmultiplication.py:53)
        [2020-03-30 21:00:17] matpy -- [ VERBOSE  ] \
                        Test Verbose Level
        [2020-03-30 21:00:17] matpy -- [  INFO    ] \
                        Initializing matrices...
        [2020-03-30 21:00:17] matpy -- [ WARNING  ] \
                        Matrix size exceeds 4 elements.
        [2020-03-30 21:00:17] matpy -- [  ERROR   ] \
                        Test Error Level (matrixmultiplication.py:54)
        [2020-03-30 21:00:17] matpy -- [ CRITICAL ] \
                        Test Critical Level (matrixmultiplication.py:55)


    These outputs are colored in the actual output but the formatting is just
    as shown above. VERBOSE is an extra added LogLevel formatting. More can be
    added below the comment `EXTRA LEVELS` in the same way the VERBOSE
    is added.

    The variable VERBOSE is given at the top of the module. That way it can be
    changed for all depending function

    """

    # Setting up the different ANSI color escape sequences for color terminals:
    # Codes are 3/4-bit codes from here:
    # https://en.wikipedia.org/wiki/ANSI_escape_code
    # For 8-colors: use "\x1b[38;5;<n>m" where <n> is the number of the color.
    grey = "\x1b[38;21m"
    green = "\x1b[38;5;64m"
    light_grey = "\x1b[38;5;240m"
    dark_blue = "\x1b[38;5;25m"
    light_blue = "\x1b[38;5;69m"
    yellow = "\x1b[33;21m"
    red = "\x1b[31;21m"
    dark_red = "\x1b[38;5;97m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"

    # Formats The spaces accommodate the different length of the words and
    # amount of detail wanted in the message:
    time_fmt = light_grey + "[%(asctime)s]" + reset
    name_fmt = "-- %(name)s -"
    pre_fmt = time_fmt + " " + name_fmt

    debug_fmt = "--- [" + light_blue + "%(levelname)s" + reset + "]:" \
                + light_blue + " %(message)s (%(filename)s:%(lineno)d)" + reset
    info_fmt = "---- [%(levelname)s]: %(message)s"
    warning_fmt = "- [" + yellow + "%(levelname)s" + reset + "]:" \
                  + yellow + " %(message)s" + reset
    error_fmt = "--- [" + red + "%(levelname)s" + reset + "]:" \
                + red + " %(message)s (%(filename)s:%(lineno)d)" + reset
    critical_fmt = " [" + bold_red + "%(levelname)s" + reset + "]:" \
                   + bold_red + " %(message)s (%(filename)s:%(lineno)d)" \
                   + reset

    # Create format dictionary
    FORMATS = {
        logging.DEBUG: pre_fmt + debug_fmt,
        logging.INFO: pre_fmt + info_fmt,
        logging.WARNING: pre_fmt + warning_fmt,
        logging.ERROR: pre_fmt + error_fmt,
        logging.CRITICAL: pre_fmt + critical_fmt
    }

    # EXTRA LEVELS
    # Verbose
    addLoggingLevel('VERBOSE', VERBOSE)
    verbose_fmt = "- [%(levelname)s]: %(message)s"
    FORMATS[logging.VERBOSE] = pre_fmt + verbose_fmt  # add to format dict

    # Initialize with a default logging.Formatter
    def __init__(self):
        super().__init__(fmt="%(levelno)d: %(msg)s", datefmt=None, style='%')

    def format(self, record):

        # Use the logging.LEVEL to get the right formatting
        log_fmt = self.FORMATS.get(record.levelno)

        # Create new formatter with modified timestamp formatting.
        formatter = logging.Formatter(log_fmt, "%Y-%m-%d %H:%M:%S")

        # Return
        return formatter.format(record)


def modify_logger(logger):
    logger.propagate = 0

    # Add formatter
    ch = logging.StreamHandler()
    ch.setFormatter(CustomFormatter())
    logger.addHandler(ch)

    return logger



#!/usr/bin/env pythoiin
# -*- coding: utf-8 -*-
"""
Calculate the adjoint source weighting based on the station and source
distribution.

:copyright:
    Wenjie Lei (lei@princeton.edu), 2016
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""
from __future__ import print_function, division, absolute_import

import os
from collections import defaultdict
import numpy as np
from pprint import pprint

from spaceweight import SpherePoint
from spaceweight import SphereDistRel
from gcmt3d.utils.io import check_dict_keys, load_json


def _receiver_validator(weights, rec_wcounts, cat_wcounts):
    """
    Validate the receiver weights, and make sure it sums to
    category window counts.

    :param weights:
    :param rec_wcounts:
    :param cat_wcounts:
    :return:
    """
    wsum = 0
    for chan, chan_weight in weights.items():
        nwin = rec_wcounts[chan]
        wsum += chan_weight * nwin

    print("Summation of (rec_weights * rec_nwins): %.2f" % wsum)
    if not np.isclose(wsum, cat_wcounts):
        raise ValueError("receiver validator fails: %f, %f" %
                         (wsum, cat_wcounts))


def calculate_receiver_window_counts(windows):
    """
    Given windows, count them based on receiver level and category
    (component) level.

    :param windows:
    :return:
    """
    rec_wcounts = defaultdict(dict)
    cat_wcounts = defaultdict(lambda: 0)
    for sta, sta_window in windows.items():
        for chan, chan_win in sta_window.items():
            comp = chan.split(".")[-1]
            nwin = len(chan_win)
            if nwin == 0:
                continue
            rec_wcounts[comp][chan] = nwin
            cat_wcounts[comp] += nwin

    return rec_wcounts, cat_wcounts


def assign_receiver_to_points(channels, stations):
    """
    Assign the receiver information to SpherePoint

    :param rec_wcounts:
    :param stations:
    :return:
    """
    points = []
    for chan in channels:
        component = chan.split(".")[-1][-1]
        if component == "Z":
            # it is important to set the default weight value to
            # 1.0 here. Becuase if there is no receiver weightings,
            # then this default weights will be used
            point = SpherePoint(stations[chan]["latitude"],
                                stations[chan]["longitude"],
                                tag=chan, weight=1.0)
        else:
            # for R and T component. In station file, there
            # are only `EN` or `12`
            echan = chan[:-1] + "E"
            chan1 = chan[:-1] + "1"
            zchan = chan[:-1] + "Z"

            if echan in stations:
                point = SpherePoint(stations[echan]["latitude"],
                                    stations[echan]["longitude"],
                                    tag=chan, weight=1.0)
            elif chan1 in stations:
                point = SpherePoint(stations[chan1]["latitude"],
                                    stations[chan1]["longitude"],
                                    tag=chan, weight=1.0)
            elif zchan in stations:
                point = SpherePoint(stations[zchan]["latitude"],
                                    stations[zchan]["longitude"],
                                    tag=chan, weight=1.0)
            else:
                raise ValueError("Can't find station information(%s)"
                                 % (chan))
        points.append(point)
    return points


def get_receiver_weights(component, center, points, max_ratio, plot=False,
                         figname_prefix=None):
    """
    Calculate the receiver weights given receiver(points) distribution.

    :param component:
    :param center:
    :param points:
    :param max_ratio:
    :param plot:
    :param figname_prefix:
    :return:
    """
    # calculate weight; otherwise, leave it as default value(1)
    weightobj = SphereDistRel(points, center=center)

    if plot:
        scan_figname = figname_prefix + ".%s.smart_scan.png" % component
    else:
        scan_figname = None

    ref_distance, cond_number = weightobj.smart_scan(
        max_ratio=max_ratio, start=0.5, gap=0.5,
        drop_ratio=0.95, plot=plot,
        figname=scan_figname)

    if plot:
        map_figname = figname_prefix + ".%s.weight.pdf" % component
        weightobj.plot_global_map(figname=map_figname, lon0=180.0)

    return ref_distance, cond_number


def normalize_receiver_weights(points, wcounts):
    """
    Normalize the receiver weights

    :param points:
    :param rec_wcounts:
    :return:
    """
    wsum = 0
    nwins_total = 0
    for point in points:
        nwin = wcounts[point.tag]
        nwins_total += nwin
        wsum += point.weight * nwin
    norm_factor = nwins_total / wsum

    weights = {}
    for point in points:
        weights[point.tag] = point.weight * norm_factor

    return weights


def determine_receiver_weighting(
        src, stations, windows, search_ratio=0.35, weight_flag=True,
        plot_flag=False, figname_prefix=None):
    """
    Given one station and window information, determine the receiver
    weighting
    In one asdf file, there are still 3 components, for example,
    ["BHR", "BHT", "BHZ"]. These three components should be treated
    indepandently and weights will be calculated independantly.

    :return: dict of weights which contains 3 components. Each components
        contains weights values
    """
    center = SpherePoint(src["latitude"], src["longitude"],
                         tag="source")

    rec_wcounts, cat_wcounts = calculate_receiver_window_counts(windows)

    weights = {}
    # in each components, calculate weight
    ref_dists = {}
    cond_nums = {}
    for comp, comp_info in rec_wcounts.items():
        print("-" * 10 + "\nComponent: %s" % comp)
        points = assign_receiver_to_points(comp_info, stations)
        print("Number of receivers: %d" % len(points))
        print("Number of windows: %d" % cat_wcounts[comp])

        if weight_flag:
            ref_dists[comp], cond_nums[comp] = \
                get_receiver_weights(comp, center, points, search_ratio,
                                     plot=plot_flag,
                                     figname_prefix=figname_prefix)
        else:
            # if not weight the receiver, then just leave the weight
            # values all to the default values(1.0). Leave the ref_dists
            # and cond_nums to None since there is no weighting
            # procedures. However, normalization still needs to be
            # applied later.
            ref_dists[comp] = None
            cond_nums[comp] = None

        # normlalize the receiver weights according to the reciever
        # window counts(normalization requirements)
        weights[comp] = normalize_receiver_weights(points, rec_wcounts[comp])

        _receiver_validator(weights[comp], rec_wcounts[comp],
                            cat_wcounts[comp])

    return {"rec_weights": weights, "rec_wcounts": rec_wcounts,
            "cat_wcounts": cat_wcounts, "rec_ref_dists": ref_dists,
            "rec_cond_nums": cond_nums}


def calculate_receiver_weights_interface(
        src_info, path_info, weighting_param, _verbose=True):
    """
    The user interface(API) for calculation the receiver weighting
    in pypaw

    :param src_info: keys contains ["latitude", "longitude"]
    :type src_info: dict
    :param path_info: keys contains ["station_file", "window_file",
        "output_file"]
    :type path_info: dict
    :param weighting_param: keys contains ["flag", "plot", "search_ratio"]
    :type weighting_param: dict
    """
    check_dict_keys(src_info, ["latitude", "longitude", "depth_in_m"])
    check_dict_keys(path_info, ["station_file", "window_file", "output_file"])
    check_dict_keys(weighting_param, ["flag", "plot", "search_ratio"])

    search_ratio = weighting_param["search_ratio"]
    plot_flag = weighting_param["plot"]
    weight_flag = weighting_param["flag"]
    # each file still contains 3-component
    if _verbose:
        print("src_info: %s" % src_info)
        print("path_info:")
        pprint(path_info)
        print("weighting param:")
        pprint(weighting_param)

    station_info = load_json(path_info["station_file"])
    window_info = load_json(path_info["window_file"])

    outputdir = os.path.dirname(path_info["output_file"])
    if not os.path.exists(outputdir):
        os.makedirs(outputdir)
    figname_prefix = os.path.join(outputdir, "weights")

    _results = determine_receiver_weighting(
        src_info, station_info, window_info,
        search_ratio=search_ratio,
        weight_flag=weight_flag,
        plot_flag=plot_flag, figname_prefix=figname_prefix)

    return _results


def _category_validator(weights, wcounts):
    """
    Validate the category weights

    :param weights:
    :param counts:
    :return:
    """
    wsum = 0.0
    nwins_total = 0
    for p, pinfo in weights.items():
        for c in pinfo:
            wsum += weights[p][c] * wcounts[p][c]
            nwins_total += wcounts[p][c]

    print("Summation of (cat_weight * cat_nwins): %.2f" % wsum)
    if not np.isclose(wsum, nwins_total):
        raise ValueError("Category validator fails: %f, %f" %
                         (wsum, nwins_total))


def normalize_category_weights(category_ratio, cat_wcounts):
    """
    """
    print("category ratio:")
    pprint(category_ratio)
    print("category window counts:")
    pprint(cat_wcounts)
    sumv = 0
    nwins_total = 0
    for p, pinfo in cat_wcounts.items():
        for c in pinfo:
            sumv += cat_wcounts[p][c] * category_ratio[p][c]
            nwins_total += cat_wcounts[p][c]

    if nwins_total == 0 or sumv == 0.0:
        normc = 1.0
    else:
        normc = nwins_total / sumv

    weights = {}
    for p, pinfo in cat_wcounts.items():
        weights[p] = {}
        for c in pinfo:
            weights[p][c] = normc * category_ratio[p][c]

    return weights


def check_category_ratio_consistency(cat_weight_ratio, cat_wcounts):
    """
    check the category weighting ratio(provide by user) has
    the same period band and component as in read seismic
    data.
    """
    err = 0
    # check consistency
    for p, pinfo in cat_weight_ratio.items():
        for c in pinfo:
            try:
                cat_wcounts[p][c]
            except KeyError:
                err = 1
                print("Missing %s.%s" % (p, c))
    if err:
        raise ValueError("category weighting ratio information is not "
                         "consistent with window information")


def calculate_category_weights_interface(category_param, cat_wcounts):
    """
    User interface(API) for calculating category weights

    Calculate the category weighting based on window counts
    in each category. The weighting ratios for different categoies
    are input parameters. So this function only normlizes
    the weights without change the ratio.
    !!! WARNING !!!
    It is the ratio between different categories, not the window
    counts. For example, the value will be mostly likely to be
    set to 1/N_w(for windows, less weights to balance categories)

    :param weight_param: user parameter, which contains the weighting
        ratio for each category.
    :type weight_param: dict
    :param cat_wcounts: category window counts, which should contains
        the same period band as category_param['ratio']
    :type cat_wcounts: dict
    """
    check_dict_keys(category_param, ["flag", "ratio"])
    check_category_ratio_consistency(category_param["ratio"], cat_wcounts)

    weights = normalize_category_weights(
        category_param["ratio"], cat_wcounts)
    _category_validator(weights, cat_wcounts)

    print("Final category weights:")
    pprint(weights)
    return weights


def combine_receiver_and_category_weights(rec_weights, cat_weights):
    """
    Combine weights for receiver weighting and category weighting
    """
    # combine weights
    weights = {}
    for period, period_info in rec_weights.items():
        weights[period] = {}
        for comp, comp_info in period_info.items():
            for chan_id in comp_info:
                rec_weight = comp_info[chan_id]
                cat_weight = cat_weights[period][comp]
                _weight = {"receiver": rec_weight,
                           "category": cat_weight}
                _weight["weight"] = \
                    rec_weight * cat_weight
                weights[period][chan_id] = _weight
    return weights



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Parent class for general asdf processing. Wraps things like MPI
and parallel I/O so they are invisible to users.

:copyright:
    Wenjie Lei (lei@princeton.edu), 2016
    Lucas Sawade (lsawade@princeton.edu) 2019
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""
from __future__ import (absolute_import, division, print_function)
import os
import warnings
from pyasdf import ASDFDataSet
from .utils import smart_read_yaml, smart_read_json, is_mpi_env
from .utils import smart_check_path, smart_remove_file, smart_mkdir


class ProcASDFBase(object):

    def __init__(self, path, param, verbose=False, debug=False):

        self.comm = None
        self.rank = None

        self.path = path
        self.param = param
        self._verbose = verbose
        self._debug = debug

    def _parse_yaml(self, content):
        """
        Parse yaml file

        :param content:
        :return:
        """
        if isinstance(content, dict):
            # already in the memory
            return content
        elif isinstance(content, str):
            return smart_read_yaml(content, mpi_mode=self.mpi_mode,
                                   comm=self.comm)
        else:
            raise ValueError("Not recogonized input: %s" % content)

    def _parse_json(self, content):
        """
        Parse json file

        :param content:
        :return:
        """
        if isinstance(content, dict):
            # already in the memory
            return content
        elif isinstance(content, str):
            return smart_read_json(content, mpi_mode=self.mpi_mode,
                                   comm=self.comm)
        else:
            raise ValueError("Not recogonized input: %s" % content)

    def _parse_path(self):
        """
        How you parse the path arugment to fit your requirements
        """
        return self._parse_yaml(self.path)

    def _parse_param(self):
        """
        How you parse the param argument to fit your requirements
        """
        return self._parse_yaml(self.param)

    def detect_env(self):
        """
        Detect environment, mpi or not

        :return:
        """
        self.mpi_mode = is_mpi_env()
        if not self.mpi_mode:
            warnings.warn("MPI environment required for parallel processing or"
                          " window selection.")
        else:
            from mpi4py import MPI
            self.comm = MPI.COMM_WORLD
            self.rank = self.comm.Get_rank()

    def print_info(self, dict_obj, title=""):
        """
        Print dict. You can use it to print out information
        for path and param

        :param dict_obj:
        :param title:
        :return:
        """
        def _print_subs(_dict, title):
            print("-"*10 + title + "-"*10)
            sorted_dict = sorted(((v, k) for v, k in _dict.items()))
            for key, value in sorted_dict:
                print("%s:  %s" % (key, value))

        if not isinstance(dict_obj, dict):
            raise ValueError("Input dict_obj should be type of dict")

        if not self.mpi_mode:
            _print_subs(dict_obj, title)
        else:
            if self.rank != 0:
                return
            _print_subs(dict_obj, title)

    def load_asdf(self, filename, mode="a"):
        """
        Load asdf file

        :param filename:
        :param mode:
        :return:
        """
        if self.mpi_mode:
            return ASDFDataSet(filename, compression=None, debug=self._debug,
                               mode=mode, mpi=self.mpi_mode)
        else:
            return ASDFDataSet(filename, mode=mode)

    def check_input_file(self, filename):
        """
        Check existance of input file. If not, raise ValueError
        """
        if not smart_check_path(filename, mpi_mode=self.mpi_mode,
                                comm=self.comm):
            raise ValueError("Input file not exists: %s" % filename)

    def check_output_file(self, filename, remove_flag=True):
        """
        Check existence of output file. If directory of output file
        not exists, raise ValueError; If output file exists, remove it
        """
        dirname = os.path.dirname(filename)
        if not smart_check_path(dirname, mpi_mode=self.mpi_mode,
                                comm=self.comm):
            print("Output dir not exists and created: %s" % dirname)
            smart_mkdir(dirname, mpi_mode=self.mpi_mode,
                        comm=self.comm)

        if smart_check_path(filename, mpi_mode=self.mpi_mode,
                            comm=self.comm):
            if remove_flag:
                if self.rank == 0:
                    print("Output file already exists and removed:%s"
                          % filename)
                else:
                    pass
                smart_remove_file(filename)

    @staticmethod
    def clean_memory(asdf_ds):
        """
        Delete asdf dataset
        """
        del asdf_ds

    @staticmethod
    def _missing_keys(necessary_keys, _dict):
        """
        Check if necessary_keys exists in _dict

        :param necessary_keys:
        :param _dict:
        :return:
        """
        if not isinstance(_dict, dict):
            raise ValueError("Input _dict must be type of dict")
        error_code = 0
        for _key in necessary_keys:
            if _key not in _dict.keys():
                print("%s must be specified in parameter file" % _key)
                error_code = 1
        if error_code:
            raise ValueError("Key values missing in paramter file")

    def _core(self, par_obj, file_obj):
        """
        Pure virtual function. Needs to be implemented in the
        child class.
        """
        raise NotImplementedError()

    def _validate_path(self, path):
        pass

    def _validate_param(self, param):
        pass

    def smart_run(self):
        """
        Job launch method

        :return:
        """

        self.detect_env()

        path = self._parse_path()
        self.print_info(path, title="Path Info")

        self._validate_path(path)

        param = self._parse_param()

        self.print_info(param, title="Param Info")
        self._validate_param(param)

        self._core(path, param)



"""

This Function contains functions to build an FDSN time series download
request from an input CMT source object and List of stations.

:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU General Public License, Version 3
    (http://www.gnu.org/copyleft/gpl.html)

"""

import os
from subprocess import Popen, PIPE, STDOUT
from ...source import CMTSource
import logging
from ...log_util import modify_logger

# Make logger
logger = logging.getLogger(__name__)
modify_logger(logger)


# Input Error
class Error(Exception):
    """Base class for exceptions in this module."""
    pass


class InputError(Error):
    """Exception raised for errors in the input.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, expression, message):
        self.expression = expression
        self.message = message
        print(self.expression + "\n\n")
        print(self.message + "\n")


class DataRequest(object):
    """

    Class that handles the building of a data request

    """

    def __init__(self, cmt=None, stationlist=None, channels=['BHZ'],
                 locations=['00'], duration=0.0, starttime_offset=0,
                 outputdir=""):
        """
        :param cmt: CMTSource object from cmt Source
        :param stationlist: list of station in format
                            `[['Net1','Sta1'],['Net2','Sta2'],etc.]`
        :param channels: list of strings containing wanted channels
        :param locations: list of seismometer locations
        :param duration: duration of the requested recording from origin time
        :param starttime_offset: time to request before the origin time, e.g.,
                                 -10 would create a request starting 10 seconds
                                 before the origin time
        :param resp_format: Response format. Either "resp" (SEED) or "pz"
                            (PolesAndZeros). Default "resp"
        :param outputdir: directory to save time series in, in a subdirectory
                          called `seismograms/`
        """

        # Check if CMT input is a CMT object
        if type(cmt) != CMTSource:
            raise InputError("dataRequest initialization",
                             "None or incorrect CMT input.")

        # CMT parameter input
        self.origin_time = cmt.cmt_time
        self.origin_latitude = cmt.latitude
        self.origin_longitude = cmt.longitude
        self.eventname = cmt.eventname

        # Check if station list is a 2D list of networks and stations
        if type(stationlist) != list:
            raise InputError("dataRequest initialization",
                             "stationlist not a list")

        # Station list parameter setup
        self.stationlist = stationlist

        # Download parameters
        self.duration = duration
        self.starttime_offset = starttime_offset
        self.starttime = self.origin_time + self.starttime_offset
        self.endtime = self.starttime + self.duration
        self.channels = channels
        self.locations = locations
        self.outputdir = outputdir

        # Name of the output directory:
        # Can't be also parsed input since it's dependent on the provided
        # filename
        if self.outputdir == "":
            self.outputdir = "."
            logger.warning("No output directory chosen. Seismograms will be ")
            logger.warning("saved in current location in subdirectory ")
            logger.warning("'seismograms/'.")
        elif not os.path.exists(self.outputdir):
            os.makedirs(self.outputdir)

    @classmethod
    def from_file(cls, cmtfname,
                  stationlistfname=os.path.join(os.path.dirname(__file__),
                                                'resources', 'stations.txt'),
                  sfstationlist=False,
                  duration=0.,
                  channels=["BHZ"],
                  locations=["00"],
                  starttime_offset=0,
                  outputdir=""):
        """Creates a downloader class from an input file
        This downloader class also needs to contain the output directory for
        the traces.


        :param cmtfname: filename of input CMTSOLUTION

        :param stationlistfname: Give list of stations. Fileformat either
            ::

                NETWORK	STATION	LAT	        LON	        ELEVATION
                IU	    ADK	    +51.882300	-176.684200	130.0


            or in specfem format. Default is None, in which
            case a specfem STATIONS list is read.

        :param sfstationlist: `boolean`, if `True` list is in specfem format.
                              Default `False`
        """

        # Name of the output directory:
        # Can't be also parsed input since it's dependent on the provided
        # filename
        if outputdir == "":
            outputdir = os.path.abspath(os.path.join(cmtfname, os.pardir))

        # First Create CMT Solution
        cmt = CMTSource.from_CMTSOLUTION_file(cmtfname)

        if stationlistfname is None:

            # Get module directory
            mod_dir, __ = os.path.split(__file__)

            # set stationlistfname to given stationlist
            stationlistfname = os.path.join(mod_dir, "resources",
                                            "STATIONS")

        # Read station list file with two columns and whitespace separation
        with open(stationlistfname, 'r') as statfile:

            if sfstationlist is False:
                # For loop to add all stations to the station list
                stationlist = []
                for line in statfile:
                    # Read stations into list of stations
                    line = line.split()

                    if line[0] == "NETWORK":
                        continue
                    else:
                        # Append [network station latitude longitude elevation]
                        # to the station list
                        stationlist.append(line)
            elif sfstationlist is True:
                stationlist = []
                for line in statfile:
                    # Read stations into list of stations
                    line = line.split()

                    newline = [line[1], line[0], line[2], line[3], line[4],
                               line[5], 0]
                    # Append the [network station latitude longitude elevation]
                    # to the station list
                    stationlist.append(newline)

        return cls(cmt=cmt,
                   stationlist=stationlist,
                   channels=channels,
                   locations=locations,
                   duration=duration,
                   starttime_offset=starttime_offset,
                   outputdir=outputdir)

    def request(self):
        """
        Writes a request file to the earthquake directory. This request file is
        then used by the download function to download.

        :returns str: Path to requestfile
        """

        # Open file for writing in the earthquake directory
        path_to_file = os.path.join(self.outputdir,
                                    'station_data', 'request.txt')
        with open(path_to_file, 'w') as requestfile:

            # Writing for each parameter overarching of all is the station
            # parameter, of course

            for station in self.stationlist:
                for location in self.locations:
                    for channel in self.channels:
                        # joining just the network   & station
                        requestfile.write(
                            " ".join([" ".join([station[0], station[1]]),
                                      location, channel,
                                      self.starttime.__str__(),
                                      self.endtime.__str__()]))
                        requestfile.write("\n")

        return path_to_file

    def download(self, selection_file="", download_log_file=""):
        """Takes self.request and starts the download of the time series data
        """

        # Create request file if none is given
        if selection_file == "":
            selection_file = "\\ ".join(self.request().split())

        # Create download log file in Earthquake directory if no other
        # directory is specified
        if download_log_file == "":
            download_log_file = os.path.join(self.outputdir,
                                             "station_data",
                                             "download_log.txt")

        # If doesn't exist, create directory for responses and seismograms
        seis_path = os.path.join(self.outputdir, "seismograms/obs/")

        if os.path.exists(seis_path) == 0:
            os.makedirs(seis_path)

        # Relative path to the resources directory
        path_to_this_file = os.path.abspath(os.path.dirname(__file__))
        path_to_script = "\\ ".join(os.path.join(path_to_this_file,
                                                 "resources").split())

        # Invoke download command depending on the response format
        with open(download_log_file, "w") as out:

            Proc = Popen(" ".join(["%s/FetchData" % path_to_script,
                                   "-l", "%s" % selection_file,
                                   "-o", "%s.mseed" %
                                   os.path.join("\\ ".join(seis_path.split()),
                                                self.eventname),
                                   "-X", "%s" %
                                   "\\ ".join(self.outputdir.split())
                                   + "/station_data/" + "station.xml"]),
                         shell=True, stdout=PIPE, stderr=STDOUT)

            for line in Proc.stdout:
                # write to standard out
                logger.verbose(line.decode('utf-8').strip())

                # write to logfile
                out.write(line.decode('utf-8'))
            Proc.wait()

    def specfem_list(self, specfemfiledir=""):
        """This method takes the station list and its content to create a
        Specfem Station recording location script list.

        The text file should have following format but not contain a header

        ::

            -------------------------------------------------------------
            | Station Network Latitude Longitude Elevation(m) Burial(m) |
            -------------------------------------------------------------

        That is the a sample line would be:

        ::

            HRV   IU   42.5064  -71.5583  200.0  0.0
            :     :       :         :       :     :

        :keyword specfemfiledir: directory where to save the STATIONS file for
                                 specfem3d_globe

        """

        # Setting the Station file name and path
        if specfemfiledir == "":
            specfemfile = self.outputdir + "/station_data/STATIONS"
        else:
            specfemfile = specfemfiledir + "/station_data/STATIONS"

        with open(specfemfile, 'w') as file:
            for k, line in enumerate(self.stationlist):
                file.write('{0:11s}{1:4s}{2:12.4f}{3:12.4f}{4:10.1f}{5:8.1f}\n'
                           .format(line[1], line[0], float(line[2]),
                                   float(line[3]), float(line[4]), 0.))
        logger.verbose(" ")
        logger.verbose("STATIONS FILE WRITTEN.")
        logger.verbose(" ")

    def __str__(self):
        """
        String that contains key download info.
        """
        return_str = "\nEarthquake Parameters\n"
        return_str += "--------------------------------------------------\n"
        return_str += "Earthquake ID: %s\n" % self.eventname
        return_str += "Origin Time: %s\n" % self.origin_time
        return_str += "Origin Latitude: %s\n" % self.origin_latitude
        return_str += "Origin Longitude: %s\n\n" % self.origin_longitude

        return_str += "Download Parameters:\n"
        return_str += "--------------------------------------------------\n"
        return_str += "Starttime: %s\n" % self.starttime
        return_str += "Endtime: %s\n" % self.endtime
        return_str += "Duration [s]: %s\n" % self.duration
        return_str += "Channels: %s\n" % self.channels
        return_str += "Locations: %s\n\n" % self.locations

        return_str += "Saving Parameters:\n"
        return_str += "--------------------------------------------------\n"
        return_str += "Output Directory: %s\n" % self.outputdir

        return return_str



""" :noindex:
Setup.py file with generic info
"""
import os
from setuptools import setup
from setuptools import find_packages
from setuptools.command.test import test as testcommand

# Utility function to read the README.md file.
# Used for the long_description.  It's nice, because now 1) we have a top levelx
# README.md file and 2) it's easier to type in the README.md file than to put a raw
# string in below ...


def read(fname):
    """From Wenjie Lei 2019"""
    try:
        return open(os.path.join(os.path.dirname(__file__), fname)).read()
    except Exception as e:
        return "Can't open %s" % fname


long_description = "%s" % read("README.md")


class PyTest(testcommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.tests")]

    def initialize_options(self):
        testcommand.initialize_options(self)
        self.pytest_args = []

    def run_tests(self):
        import pytest
        import sys
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


setup(
    name="gcmt3d",
    description="Global 3D Centroid Moment Tensor Inversion",
    long_description=long_description,
    version="0.0.2",
    author="Lucas Sawade",
    author_email="lsawade@princeton.edu",
    license='GNU Lesser General Public License, Version 3',
    keywords="Global CMT, Inversion, Moment Tensor",
    url='https://github.com/lsawade/GCMT3D',
    packages=find_packages(exclude=['*.Notebooks', '*.notebooks.*',
                                    'notebooks.*', 'notebooks']),
    package_dir={"": "."},
    include_package_data=True,
    exclude_package_data={'': ['notebooks']},
    package_data={'gcmt3d': ['data/download/resources/Fetchdata',
                             'data/download/resources/stations.txt',
                             'data/management/STATIONS'],
    },
    install_requires=['numpy', 'matplotlib', 'flake8', 'obspy==1.1.1',
                      'PyYAML', 'pyflex', 'h5py', 'mpi4py', 'matplotlib',
                      'pyasdf', 'spaceweight', 'pycmt3d',
                      ],
    tests_require=['pytest'],
    cmdclass={'tests': PyTest},
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
        ("License :: OSI Approved "
         ":: GNU General Public License v3 or later (GPLv3+)"),
    ],
    #install_requires=parse_requirements("requirements.txt"),
    extras_require={
        "docs": ["sphinx", "sphinx_rtd_theme"],
        "tests": ["pytest", "py"]
    },
    entry_points={
        'console_scripts': [
            'convert2asdf = gcmt3d.bins.convert_to_asdf:main',
            'convert2sac = gcmt3d.bins.convert_to_sac:main',
            'count-windows = gcmt3d.bins.count_overall_windows:main',
            'compute-stats = gcmt3d.bins.compute_stats:main',
            'create-entry = gcmt3d.bins.create_entry:main',
            'create-inversion-dicts = gcmt3d.bins.write_inversion_dicts:main',
            'create-gridsearch-dicts = gcmt3d.bins.write_grid_search_dicts:main',
            'create-path-files = gcmt3d.bins.create_path_files:main',
            'extract-station-info = gcmt3d.bins.extract_station_info:main',
            'filter-windows = gcmt3d.bins.filter_windows:main',
            'generate-stations-asdf = gcmt3d.bins.generate_stations_asdf:main',
            'inversion = gcmt3d.bins.inversion:main',
            'gridsearch = gcmt3d.bins.gridsearch:main',
            'process-asdf = gcmt3d.bins.process_asdf:main',
            'request-data = gcmt3d.bins.request_data:main',
            'select-windows = gcmt3d.bins.window_selection_asdf:main',
            'write-sources = gcmt3d.bins.write_sources:main',
    ]}
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Methods that handles writing windows

:copyright:
    Wenjie Lei (lei@princeton.edu), 2016
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""
import json
import obspy
import yaml
import pyflex
import numpy as np


def load_window_config_yaml(filename):
    """
    Load yaml and setup pyflex.Config object

    :param filename:
    :return:
    """
    with open(filename) as fh:
        data = yaml.load(fh, Loader=yaml.FullLoader)

    if data["min_period"] > data["max_period"]:
        raise ValueError("min_period is larger than max_period in config "
                         "file: %s" % filename)

    return pyflex.Config(**data)


def write_txtfile(windows, filename):
    """
    Write windows to text file, for a list of windows. Notice that
    this method only works on LISTS of windows, for example,
    windows on a pair of traces.

    :param windows: list of windows(from same observed and synthetic)
    :type windows: list
    :param filename: output filename
    :type filename: str
    :return:
    """
    with open(filename, 'w') as fh:
        fh.write("%s\n" % windows[0].channel_id)
        fh.write("%d\n" % len(windows))
        for win in windows:
            fh.write("%10.2f %10.2f %10.2f %10.3f %10.3f\n"
                     % (win.relative_starttime, win.relative_endtime,
                        win.cc_shift, win.dlnA, win.max_cc_value))


def get_json_content(window, simple_mode=True):
    """
    Extract information from json to a dict

    :param window:
    :return:
    """
    # to be comptabile with olde pyflex, which doesn't has
    # channel_id_2. If not, assign it with "UNKNOWN"
    info = {
        "left_index": window.left,
        "right_index": window.right,
        "center_index": window.center,
        "channel_id": window.channel_id,
        "time_of_first_sample": window.time_of_first_sample,
        "max_cc_value":  window.max_cc_value,
        "cc_shift_in_samples":  window.cc_shift,
        "cc_shift_in_seconds":  window.cc_shift_in_seconds,
        "dlnA":  window.dlnA,
        "dt": window.dt,
        "min_period": window.min_period,
        "absolute_starttime": window.absolute_starttime,
        "absolute_endtime": window.absolute_endtime,
        "relative_starttime": window.relative_starttime,
        "relative_endtime": window.relative_endtime,
        "window_weight": window.weight}

    if not simple_mode:
        info["phase_arrivals"] = window.phase_arrivals

    if "channel_id_2" in dir(window):
        info["channel_id_2"] = window.channel_id_2

    return info


class WindowEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, obspy.UTCDateTime):
            return str(obj)
        # Numpy objects also require explicit handling.
        elif isinstance(obj, np.int64):
            return int(obj)
        elif isinstance(obj, np.int32):
            return int(obj)
        elif isinstance(obj, np.float64):
            return float(obj)
        elif isinstance(obj, np.float32):
            return float(obj)
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def write_jsonfile(windows, filename):
    """
    Write windows to a json file. Also, this requires windows to be
    type of list.

    :param windows: list of windows
    :param filename: output filename
    :return:
    """

    win_json = [get_json_content(_i) for _i in windows]
    with open(filename, 'w') as fh:
        j = json.dumps(win_json, cls=WindowEncoder, sort_keys=True,
                       indent=2, separators=(',', ':'))
        try:
            fh.write(j)
        except TypeError:
            fh.write(j.encode())



"""
# functions mainly for pypaw.
"""
import numpy as np
from gcmt3d.utils.io import dump_json


def sort_windows_on_channel_and_location(sta_win):
    """
    functions for merge_instruments_window. the windows
    will be sorted based on [chan][location]

    :param sta_win:
    :return:
    """
    sort_dict = {}
    for trace_id, trace_win in sta_win.items():
        chan = trace_id.split('.')[-1][0:2]
        loc = trace_id.split('.')[-2]
        if chan not in sort_dict:
            sort_dict[chan] = {}
        if loc not in sort_dict[chan]:
            sort_dict[chan][loc] = {"traces": [], "nwins": 0}
        sort_dict[chan][loc]["traces"].append(trace_id)
        sort_dict[chan][loc]["nwins"] += len(trace_win)

    # sort to trace names in order
    for chan, chan_info in sort_dict.items():
        for loc, loc_info in chan_info.items():
            loc_info['traces'] = sorted(loc_info['traces'])

    return sort_dict


def pick_location_with_more_windows(sort_dict):
    choosen = {}
    for chan, chan_info in sort_dict.items():
        if len(chan_info.keys()) == 0:
            continue

        # if multiple locations available, choose the
        # one with most number of windows
        _locs = []
        _nwins = []
        for loc, loc_info in chan_info.items():
            _locs.append(loc)
            _nwins.append(loc_info["nwins"])
        _max_idx = np.array(_nwins).argmax()
        choosen[chan] = _locs[_max_idx]

    return choosen


def merge_instruments_window(sta_win):
    """
    Merge windows from the same channel, for example, if
    there are windows from "00.BH*" and "10.BH*", keep only one
    with the most windows. For example, if "00.BH*" has 10
    windows and "10.BH*" has 20 windows, we will keep the
    "10.BH*" since it has more windows.
    """
    if len(sta_win) == 0:
        return sta_win

    sort_dict = sort_windows_on_channel_and_location(sta_win)
    choosen_locs = pick_location_with_more_windows(sort_dict)

    choosen_wins = {}
    for chan, loc in choosen_locs.items():
        trace_list = sort_dict[chan][loc]["traces"]
        for tr_id in trace_list:
            choosen_wins[tr_id] = sta_win[tr_id]

    return choosen_wins


def sort_windows_on_channel(sta_win):
    """
    Group windows from one station into channels and count
    the number of windows in that channel
    :param sta_win:
    :return:
    """
    sort_dict = {}
    for trace_id, trace_win in sta_win.items():
        chan = trace_id.split(".")[-1][0:2]
        if chan not in sort_dict:
            sort_dict[chan] = {"traces": [], "nwins": 0}

        sort_dict[chan]["traces"].append(trace_id)
        sort_dict[chan]["nwins"] += len(trace_win)

    return sort_dict


def pick_channel_with_more_windows(sort_dict):
    max_wins = -1
    max_chan = None
    for chan, chaninfo in sort_dict.items():
        if chaninfo["nwins"] > max_wins:
            max_wins = chaninfo["nwins"]
            max_chan = chan
    return max_chan


def merge_channels_window(sta_win):
    """
    Merge windows from different channels.
    This step should be done after merge instruments windows
    because after that there will only one instrument left
    on one channel.
    For example, if we have "BH" channel with 20 windows and
    "LH" has 10 windows, we will keep only the "BH" channel.
    """
    if len(sta_win) == 0:
        return sta_win

    sort_dict = sort_windows_on_channel(sta_win)
    choosen_chan = pick_channel_with_more_windows(sort_dict)
    choosen_traces = sort_dict[choosen_chan]["traces"]

    choosen_wins = {}
    for _trace_id in choosen_traces:
        choosen_wins[_trace_id] = sta_win[_trace_id]

    return choosen_wins


def merge_station_windows(windows):
    """
    Merge windows for one station.
    For example, you may have "00.BH", "10.BH", "00.LH", "10.LH" from
    different locations and channels. You may only want to keep one
    at the very end. So:

    1. select locations: keep only one location with the most
       number of windows for one channel. For example, in "00.BH" and
       "10.BH" we only keep "10.BH".
    2. select channel: after last step, we keep only "00.BH" and "10.LH",
       but ultimately we only want to keep one. So we may only choose
       "00.BH" since it has more windows.

    """
    w = merge_instruments_window(windows)
    w = merge_channels_window(w)
    return w


def merge_windows(windows):
    """
    Merge the windows(from one event, multiple stations)
    """
    new_windows = {}
    for sta, sta_info in windows.items():
        if sta_info is None:
            continue
        # merge the windows for each station
        new_windows[sta] = merge_station_windows(sta_info)
    return new_windows


def generate_log_content(windows):
    overall_log = {"stations": 0, "stations_with_windows": 0,
                   "windows": 0, "traces": 0, "traces_with_windows": 0}
    comp_log = {}
    for sta_name, sta_win in windows.items():
        if sta_win is None:
            continue
        nwin_sta = 0
        ntraces_with_windows = 0
        for trace_id, trace_win in sta_win.items():
            comp = trace_id.split(".")[-1]
            if comp not in comp_log:
                comp_log[comp] = {
                    "windows": 0, "traces": 0, "traces_with_windows": 0}
            comp_log[comp]["windows"] += len(trace_win)
            if len(trace_win) > 0:
                comp_log[comp]["traces_with_windows"] += 1
                ntraces_with_windows += 1
            comp_log[comp]["traces"] += 1
            nwin_sta += len(trace_win)

        overall_log["stations"] += 1
        overall_log["windows"] += nwin_sta
        overall_log["traces"] += len(sta_win)
        overall_log["traces_with_windows"] += ntraces_with_windows
        if nwin_sta > 0:
            overall_log["stations_with_windows"] += 1

    log = {"component": comp_log, "overall": overall_log}
    return log


def stats_all_windows(windows, obsd_tag, synt_tag,
                      instrument_merge_flag,
                      output_file):
    """
    Generate window statistic information
    """
    log = {"obsd_tag": obsd_tag, "synt_tag": synt_tag,
           "instrument_merge_flag": instrument_merge_flag}

    window_log = generate_log_content(windows)
    log.update(window_log)

    print(output_file)

    print("Windows statistic log file: %s" % output_file)
    dump_json(log, output_file)


def convert_to_window_json_to_txt(windowjson, outputtxt):
    """ Converts window

    :param windowjson:
    :param outputtxt:
    :return:
    """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Conversion utilities for different seismic datasets.

:copyright:
    Wenjie Lei (lei@princeton.edu), 2016
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""
from __future__ import (absolute_import, division, print_function)
import os
import glob
import numpy as np
from obspy import UTCDateTime, read
from ..station.utils import create_simple_inventory
from pyasdf import ASDFDataSet
from .utils import smart_read_yaml, drawProgressBar, timing


def add_waveform_to_asdf(ds, waveform_filelist, tag, event=None,
                         create_simple_inv=False, status_bar=False):

    nwaveform = len(waveform_filelist)
    sta_dict = {}
    # Add waveforms.
    for _i, filename in enumerate(waveform_filelist):
        if not os.path.exists(filename):
            raise ValueError("File not exist %i of %i: %s"
                             % (_i, nwaveform, filename))

        try:
            st = read(filename)
            ds.add_waveforms(st, tag=tag, event_id=event)
        except Exception as err:
            print("Error converting(%s) due to: %s" % (filename, err))
            continue
        if create_simple_inv:
            for tr in st:
                sta_tag = "%s_%s" % (tr.stats.network, tr.stats.station)
                if sta_tag not in sta_dict.keys():
                    try:
                        _sac = tr.stats.sac
                    except Exception:
                        raise ValueError("The original data format should be"
                                         "sac format to extract station"
                                         "information")
                    sta_dict[sta_tag] = [tr.stats.network, tr.stats.station,
                                         _sac["stla"], _sac["stlo"],
                                         _sac["stel"], _sac["stdp"]]
                else:
                    continue

        if status_bar:
            drawProgressBar((_i+1)/nwaveform, "Adding Waveform data")
    return sta_dict


def add_stationxml_to_asdf(ds, staxml_filelist, event=None,
                           create_simple_inv=False, sta_dict=None,
                           status_bar=False):
    # Add StationXML files.
    if create_simple_inv:
        if event is None:
            start_date = UTCDateTime.now()
        else:
            origin = event.preferred_origin() or event.origins[0]
            event_time = origin.time
            start_date = event_time - 300.0
        nstaxml = len(sta_dict)
        count = 0
        for tag, value in sta_dict.items():
            count += 1
            inv = create_simple_inventory(
                value[0], value[1], latitude=value[2], longitude=value[3],
                elevation=value[4], depth=value[5], start_date=start_date)
            ds.add_stationxml(inv)
            if status_bar > 0:
                drawProgressBar((count)/nstaxml,
                                "Adding StationXML(created) data")
    else:
        nstaxml = len(staxml_filelist)
        if staxml_filelist is not None and nstaxml > 0:
            for _i, filename in enumerate(staxml_filelist):
                if not os.path.exists(filename):
                    raise ValueError("Staxml not exist %i of %i: %s"
                                     % (_i, nstaxml, filename))
                try:
                    ds.add_stationxml(filename)
                except Exception as err:
                    print("Error convert(%s) due to:%s" % (filename, err))
                if status_bar > 0:
                    drawProgressBar((_i+1)/nstaxml, "Adding StationXML data")
        else:
            print("No stationxml added")


@timing
def convert_to_asdf(asdf_fn, waveform_filelist, tag, quakemlfile=None,
                    staxml_filelist=None, verbose=False, status_bar=False,
                    create_simple_inv=False):
    """
    Convert files(sac or mseed) to asdf
    """

    if verbose:
        print("*"*10 + " ASDF Converter " + "*"*10)

    nwaveform = len(waveform_filelist)
    if nwaveform == 0:
        print("No file specified. Return...")
        return
    if os.path.exists(asdf_fn):
        raise Exception("File '%s' exists." % asdf_fn)

    ds = ASDFDataSet(asdf_fn, mode='a')

    # Add event
    if quakemlfile:
        if not os.path.exists(quakemlfile):
            raise ValueError("Quakeml file not exists:%s" % quakemlfile)
        ds.add_quakeml(quakemlfile)
        event = ds.events[0]
        if status_bar:
            drawProgressBar(1.0, "Adding Quakeml data")
    else:
        raise ValueError("No Event file")

    sta_dict = add_waveform_to_asdf(ds, waveform_filelist, tag, event=event,
                                    create_simple_inv=create_simple_inv,
                                    status_bar=status_bar)

    add_stationxml_to_asdf(ds, staxml_filelist, event=event,
                           create_simple_inv=create_simple_inv,
                           sta_dict=sta_dict,
                           status_bar=status_bar)

    if verbose:
        print("ASDF filesize: %s" % ds.pretty_filesize)
    del ds


def write_stream_to_sac(stream, outputdir, tag=""):
    for tr in stream:
        if tag == "":
            filename = os.path.join(outputdir, "%s.sac" % tr.id)
        else:
            filename = os.path.join(outputdir, "%s.%s.sac" % (tr.id, tag))
        tr.write(filename, format="SAC")


@timing
def convert_from_asdf(asdf_fn, outputdir, tag=None, filetype="sac",
                      output_staxml=True, output_quakeml=True,
                      _verbose=True):
    """
    Convert the waveform in asdf to different types of file
    """
    filetype = filetype.upper()
    if filetype not in ["SAC", "MSEED"]:
        raise ValueError("Supported filetype: 1) sac; 2) mseed")

    if not os.path.exists(asdf_fn):
        raise ValueError("No asdf file: %s" % asdf_fn)
    if not os.path.exists(outputdir):
        os.makedirs(outputdir)

    if isinstance(tag, str):
        tag_list = [tag]

    print("Input ASDF: %s" % asdf_fn)
    print("Output dir: %s" % outputdir)
    print("Output StationXML and Quakeml: [%s, %s]" % (output_staxml,
                                                       output_quakeml))

    ds = ASDFDataSet(asdf_fn, mode='r')

    if output_quakeml:
        if len(ds.events) >= 1:
            filename = os.path.join(outputdir, "Quakeml.xml")
            if _verbose:
                print("Quakeml file: %s" % filename)
            ds.events.write(filename, format="QUAKEML")

    sta_list = ds.waveforms.list()

    for station_name in sta_list:
        if _verbose:
            print("Convert station: %s" % station_name)
        station_name2 = station_name.replace(".", "_")
        station = getattr(ds.waveforms, station_name2)
        default_tag_list = station.get_waveform_tags()
        if tag is None:
            tag_list = default_tag_list
        for _tag in tag_list:
            if _tag not in default_tag_list:
                print("Tag(%s) not in Station(%s) taglist(%s)" %
                      (_tag, station_name2, default_tag_list))
                continue
            try:
                stream, inv = ds.get_data_for_tag(station_name2, _tag)
            except Exception:
                print("Error for station:", station_name2)
            if filetype == "SAC":
                write_stream_to_sac(stream, outputdir, _tag)
            elif filetype == "MSEED":
                filename = os.path.join(outputdir, "%s.%s.mseed"
                                        % (station_name, _tag))
                stream.write(filename, format="MSEED")
            if output_staxml:
                filename = os.path.join(outputdir, "%s.%s.xml"
                                        % (station_name, _tag))
                try:
                    inv.write(filename, format="STATIONXML")
                except Exception:
                    print("Error creating STATIONXML: %f" % filename)

    del ds


@timing
def convert_adjsrcs_from_asdf(asdf_fn, outputdir, _verbose=True):
    """
    Convert adjoint sources from asdf to ASCII file(for specfem3d_globe use)
    """
    if not os.path.exists(asdf_fn):
        raise ValueError("No asdf file: %s" % asdf_fn)
    if not os.path.exists(outputdir):
        os.makedirs(outputdir)

    print("Input ASDF: %s" % asdf_fn)
    print("Output dir: %s" % outputdir)

    ds = ASDFDataSet(asdf_fn, mode='r')
    if "AdjointSources" not in ds.auxiliary_data:
        print("No adjoint source exists in asdf file: %s" % asdf_fn)
        return
    adjsrcs = ds.auxiliary_data.AdjointSources
    nadj = len(adjsrcs)
    print("Number of adjoint sources: %d" % nadj)

    # get event time
    origin = ds.events[0].preferred_origin()
    eventtime = origin.time

    for idx, adj in enumerate(adjsrcs):
        if _verbose:
            print("Adjoint sources(%d/%d) from: %s" % (idx, nadj, adj.path))
        trace_starttime = UTCDateTime(adj.parameters["starttime"])
        time_offset = trace_starttime - eventtime

        dt = adj.parameters['dt']
        npts = len(adj.data)
        times = np.array([time_offset + i * dt for i in range(npts)])
        _data = np.zeros([npts, 2])
        _data[:, 0] = times[:]
        _data[:, 1] = adj.data[:]
        adj_path = adj.path.replace("_", ".")
        filename = os.path.join(outputdir, "%s.adj" % adj_path)
        np.savetxt(filename, _data)


class ConvertASDF(object):

    def __init__(self, path, verbose=False, status_bar=False):
        self.path = path
        self._verbose = verbose
        self._status_bar = status_bar

    @staticmethod
    def print_info(waveform_files, tag, staxml_files, quakemlfile,
                   output_fn, create_simple_inv):

        waveform_dir = set([os.path.dirname(_i) for _i in waveform_files])

        print("-"*20)
        print("Quakeml files: ", quakemlfile)
        print("Waveform dirs:", waveform_dir)
        print("Number of waveform files:", len(waveform_files))
        if not create_simple_inv:
            staxml_dir = set([os.path.dirname(_i) for _i in staxml_files])
            print("Stationxml dirs:", staxml_dir)
            print("Number of Stationxml files:", len(staxml_files))
        else:
            print("Stationxml files: generated based on input waveform")
        print("Output filename:", output_fn)

    @staticmethod
    def clean_output(output_fn):
        basepath = os.path.dirname(output_fn)
        if not os.path.exists(basepath):
            print("Output dir not exists so created: %s" % basepath)
            os.makedirs(basepath)
        if os.path.exists(output_fn):
            print("Outfile exist and being removed:", output_fn)
            os.remove(output_fn)

    @staticmethod
    def _parse_path(path):

        if "waveform_files" in path:
            waveformfiles = glob.glob(path["waveform_files"])
            filetype = None
        elif "waveform_dir" in path:
            waveformdir = path["waveform_dir"]
            filetype = path["filetype"].lower()
            waveform_pattern = os.path.join(waveformdir, "*"+filetype)
            waveformfiles = glob.glob(waveform_pattern)
        else:
            raise ValueError("missing keywords in yaml file, 'waveform_files'"
                             "or 'waveformdir'")

        if "staxml_files" in path:
            staxmlfiles = glob.glob(path["staxml_files"])
            create_simple_inv = False
        elif "staxml_dir" in path:
            staxmldir = path["staxml_dir"]
            staxml_pattern = os.path.join(staxmldir, '*.xml')
            staxmlfiles = glob.glob(staxml_pattern)
            create_simple_inv = False
        else:
            if filetype != "sac":
                raise ValueError("create_simple_inv only supports sac")
            staxmlfiles = None
            create_simple_inv = True

        tag = path["tag"]
        quakemlfile = path["quakeml_file"]
        outputfile = path["output_file"]

        return waveformfiles, tag, staxmlfiles, quakemlfile, outputfile,\
            create_simple_inv

    def _run_subs(self, path):

        waveformfiles, tag, staxmlfiles, quakemlfile, outputfile, \
            create_simple_inv = self._parse_path(path)

        self.print_info(waveformfiles, tag, staxmlfiles, quakemlfile,
                        outputfile, create_simple_inv)

        self.clean_output(outputfile)

        convert_to_asdf(outputfile, waveformfiles, tag,
                        quakemlfile=quakemlfile,
                        staxml_filelist=staxmlfiles,
                        verbose=self._verbose, status_bar=self._status_bar,
                        create_simple_inv=create_simple_inv)

    def run(self):
        path = smart_read_yaml(self.path, mpi_mode=False)
        if isinstance(path, list):
            for _path in path:
                self._run_subs(_path)
        else:
            print(path)
            self._run_subs(path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This is script is used for stats the overall window counts on all window
files.

Suppose you have 1,000 events and 3 period bands, then the number of window
files will be about 3,000. Next, you want to know how many windows do
each category have and the number will then be used in the weighting
stage.
If "weight_output_file" is in the path file, then a default weight param
file will also be generated. Be sure to modify some values and then it can
be used in the weighting stage.
"""
from __future__ import print_function, division, absolute_import
import numpy as np
import argparse
from pprint import pprint
from gcmt3d.window.utils import generate_log_content
from gcmt3d.asdf.utils import load_json, load_yaml, dump_json, dump_yaml


def stats_one_window_file(filename):
    """
    Given one window file, return the window counts for different
    component
    """
    windows = load_json(filename)
    log = generate_log_content(windows)
    results = {}
    for comp in log["component"]:
        results[comp] = log["component"][comp]["windows"]
    return results


def update_overall_wcounts(overall_wcounts, period, one_file_results):
    """
    Given the window counts from one file, update the overall
    window counts
    """
    if period not in overall_wcounts:
        overall_wcounts[period] = {}
    for comp in one_file_results:
        if comp == "total":
            # skip the total, which is not a component
            continue
        if comp not in overall_wcounts[period]:
            overall_wcounts[period][comp] = 0
        overall_wcounts[period][comp] += one_file_results[comp]


def _validate_ratio(overall_wcounts, ratio):
    prods = []
    for p, pinfo in overall_wcounts.iteritems():
        for c, cinfo in pinfo.iteritems():
            prods.append(cinfo * ratio[p][c])

    if not all([np.isclose(prods[0], _v) for _v in prods]):
        print("Failed wcounts and ratio validator!")


def ensemble_default_weight_param_file(overall_wcounts, param, outputfile):
    """
    Ensemble a default weight file based on overall window counts
    and weight ratio provided by the user
    """
    maxv = 0
    # find max first
    for p, pinfo in overall_wcounts.iteritems():
        for c, cinfo in pinfo.iteritems():
            if cinfo > maxv:
                maxv = cinfo

    raw_ratio = {}
    ratio = {}
    for p, pinfo in overall_wcounts.iteritems():
        ratio[str(p)] = {}
        raw_ratio[str(p)] = {}
        for c, cinfo in pinfo.iteritems():
            _user_weight = param["user_weight_ratio"][p][c]
            # the raw ratio without user-definded weight
            # this is essentially window counts ratio
            raw_ratio[str(p)][str(c)] = maxv / cinfo
            # the true weight ratio used
            ratio[str(p)][str(c)] = _user_weight * (maxv / cinfo)

    _validate_ratio(overall_wcounts, raw_ratio)
    print("===> Raw window counts ratio:")
    pprint(raw_ratio)
    print("===> Modified weight ratio:")
    pprint(ratio)

    default_content = {
        "receiver_weighting": {
            "flag": True, "plot": False, "search_ratio": "TO BE FILLED"},
        "category_weighting": {
            "flag": True, "ratio": ratio}
    }
    dump_yaml(default_content, outputfile)


def stats_all_window_file(path, _verbose):
    detailed_event_windows = {}
    overall_wcounts = {}

    print("=" * 10 + " Start counting windows " + "=" * 10)
    eventlist = path["input"].keys()
    eventlist.sort()
    nevents = len(eventlist)
    for idxe, e in enumerate(eventlist):
        period_bands = path["input"][e].keys()
        period_bands.sort()
        print("-" * 8 + "[%d/%d]%s" % (idxe, nevents, e)
              + "-" * 8)
        detailed_event_windows[e] = {}
        event_total = 0
        for p in period_bands:
            one_file_results = stats_one_window_file(path["input"][e][p])
            print("period[%s]: %s" % (p, one_file_results))
            detailed_event_windows[e][p] = one_file_results
            period_total = 0
            for comp, comp_counts in one_file_results.iteritems():
                period_total += comp_counts
            detailed_event_windows[e][p]["total"] = period_total
            event_total += period_total
            # update for overall window counts
            update_overall_wcounts(overall_wcounts, p, one_file_results)
        detailed_event_windows[e]["total"] = event_total

    # dump results
    print("=" * 10 + " dump results " + "=" * 10)
    outputfile = path["output_file"]
    content = {"detailed_information": detailed_event_windows,
               "summary": overall_wcounts}
    print("outputfile: %s" % outputfile)
    dump_json(content, outputfile)

    return overall_wcounts


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', action='store', dest='path_file',
                        required=True, help='path file')
    parser.add_argument('-p', action='store', dest='param_file',
                        required=True, help='param file')
    parser.add_argument('-v', action='store_true', dest='verbose',
                        help='verbose flag')
    args = parser.parse_args()

    path = load_json(args.path_file)
    print("=" * 10 + " Path information " + "=" * 10)
    pprint(path)

    param = load_yaml(args.param_file)
    print("=" * 10 + " Param information " + "=" * 10)
    pprint(param)

    overall_wcounts = stats_all_window_file(path, args.verbose)

    if "weight_output_file" in path:
        print("weight default param file: %s" % path["weight_output_file"])
        ensemble_default_weight_param_file(
            overall_wcounts, param, path["weight_output_file"])


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This script will download the observed data. To the necessary places.


:copyright:
    Lucas Sawade (lsawade@princeton.edu)
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""

from gcmt3d.utils.download import download_waveform
from gcmt3d.utils.download import download_stationxml
from gcmt3d.utils.download import read_station_file
from gcmt3d.asdf.utils import smart_read_yaml, is_mpi_env
from gcmt3d.source import CMTSource
import os


def data_request(cmt_filename):
    # Set directories of the parameter files
    param_path = os.path.join(os.path.dirname(
        os.path.dirname(os.path.abspath(__file__))), "params")

    request_param_path = os.path.join(param_path,
                                      "RequestParams/RequestParams.yml")

    # Read the parameter file
    rCparams = smart_read_yaml(request_param_path, mpi_mode=is_mpi_env())

    # Earthquake and Station parameters
    cmt_dir = os.path.dirname(cmt_filename)
    station_dir = os.path.join(cmt_dir, "station_data")

    # Get STATIONS file from CMT directory
    stationsfile = os.path.join(station_dir, "STATIONS")

    # Observed output dir
    obsd_dir = os.path.join(cmt_dir, "seismograms", "obs")

    # CMT parameter input
    cmt = CMTSource.from_CMTSOLUTION_file(cmt_filename)
    duration = rCparams['duration']
    starttime_offset = rCparams['starttime_offset']

    starttime = cmt.origin_time + starttime_offset
    endtime = starttime + duration

    # Get station_list from station_file in database entry
    stations = read_station_file(stationsfile)
    station_ids = [station[0] + "_" + station[1]
                   for station in stations]

    # Download Station Data
    _, _, filtered_station_ids = \
        download_stationxml(station_ids, starttime, endtime,
                            outputdir=station_dir, client=None,
                            level="response")

    # Download waveform
    download_waveform(filtered_station_ids, starttime, endtime,
                      outputdir=obsd_dir, client=None)



#!/usr/bin/env python
"""
Scripts that generate stations file from asdf file.
1) If there are stations in waveforms, then a file `STATIONS`
will be generated.
2) if there are stations in AuxlilaryData.AdjointSource,
then a file `STATIONS_ADJOINT` will be generated.

The output STATIONS file follows the format in SPECFEM3D_GLOBE.

:copyright:
    Wenjie Lei (lei@princeton.edu), 2016
:license:
    GNU General Public License, Version 3
    (http://www.gnu.org/copyleft/gpl.html)
"""
from __future__ import (print_function, division)

import argparse
import os

import pyasdf

from gcmt3d.asdf.stations import extract_adjoint_stations
from gcmt3d.asdf.stations import extract_waveform_stations
from gcmt3d.station.utils import write_stations_file


def generate_waveform_stations(asdf, outputfn):
    print("Input asdf: %s" % asdf)
    sta_dict = extract_waveform_stations(asdf)

    print("Output file: %s" % outputfn)
    print("Number of stations: %d" % len(sta_dict))
    if len(sta_dict) > 0:
        write_stations_file(sta_dict, outputfn)


def generate_adjoint_stations(asdf, outputfn):
    print("Input asdf: %s" % asdf)
    sta_dict = extract_adjoint_stations(asdf)

    print("Output file: %s" % outputfn)
    print("Number of stations: %d" % len(sta_dict))
    if len(sta_dict) > 0:
        write_stations_file(sta_dict, outputfn)


def main():
    parser = argparse.ArgumentParser(
        description="Generate SPECFEM-like STATION file")
    parser.add_argument('-o', action='store', dest="outputfn",
                        default="STATIONS",
                        help="output station filename prefix")
    parser.add_argument('filename', help="Input ASDF filename")

    args = parser.parse_args()
    if not os.path.exists(args.filename):
        raise ValueError("Input file not exists: %s" % args.filename)

    ds = pyasdf.ASDFDataSet(args.filename, mode='r')
    generate_waveform_stations(ds, args.outputfn)

    adjoint_fn = args.outputfn + "_ADJOINT"
    generate_adjoint_stations(ds, adjoint_fn)


if __name__ == '__main__':
    main()



from __future__ import print_function, division, absolute_import

# from .process import ProcASDF       # NOQA
# from .window import WindowASDF      # NOQA
# from .convert import ConvertASDF, convert_from_asdf   # NOQA
# from .convert import convert_adjsrcs_from_asdf        # NOQA



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Parent class for signal processing asdf file and
handles parallel I/O so they are invisible to users.

:copyright:
    Wenjie Lei (lei@princeton.edu), 2016
:license:
    GNU Lesser General Public License, version 3 (LGPLv3)
    (http://www.gnu.org/licenses/lgpl-3.0.en.html)
"""
from __future__ import (print_function, division, absolute_import)
import inspect
from functools import partial
from ..signal.process import process_stream, add_stream  # NOQA
from .procbase import ProcASDFBase


def check_param_keywords(param):
    """
    Check the param keywords are the same with the keywords list of
    the function of process_stream
    """
    default_param = inspect.getfullargspec(process_stream).args
    default_param.remove("st")
    default_param.remove("inventory")
    if not param["remove_response_flag"]:
        # water_level is only used in remove instrument response
        default_param.remove("water_level")
    if set(default_param) != set(param.keys()):
        print("Missing: %s" % (set(default_param) - set(param.keys())))
        print("Redundant: %s" % (set(param.keys()) - set(default_param)))
        raise ValueError("Param is not consistent with function argument list")


def process_wrapper(stream, inv, param=None):
    """
    Process function wrapper for pyasdf

    :param stream:
    :param inv:
    :param param:
    :return:
    """
    param["inventory"] = inv
    return process_stream(stream, **param)


def update_param(event, param):
    """ update the param based on event information """
    origin = event.preferred_origin()
    origin = event.preferred_origin()
    event_latitude = origin.latitude
    event_longitude = origin.longitude
    event_time = origin.time

    # figure out interpolation parameter
    param["starttime"] = event_time + param["relative_starttime"]
    param.pop("relative_starttime")
    param["endtime"] = event_time + param["relative_endtime"]
    param.pop("relative_endtime")
    param["event_latitude"] = event_latitude
    param["event_longitude"] = event_longitude


class ProcASDF(ProcASDFBase):

    def __init__(self, path, param, verbose=False, debug=False):
        ProcASDFBase.__init__(self, path, param, verbose=verbose,
                              debug=debug)

    def _validate_path(self, path):
        necessary_keys = ["input_asdf", "input_tag",
                          "output_asdf", "output_tag"]

        self._missing_keys(necessary_keys, path)

    def _validate_param(self, param):
        necessary_keys = ("remove_response_flag", "filter_flag", "pre_filt",
                          "relative_starttime", "relative_endtime",
                          "resample_flag", "sampling_rate", "rotate_flag",
                          "sanity_check")

        self._missing_keys(necessary_keys, param)

    def _core(self, path, param):

        input_asdf = path["input_asdf"]
        input_tag = path["input_tag"]
        output_asdf = path["output_asdf"]
        output_tag = path["output_tag"]

        self.check_input_file(input_asdf)
        self.check_output_file(output_asdf, remove_flag=True)

        # WJ: set to 'a' for now since SPECFEM output is
        # a incomplete asdf file, missing the "auxiliary_data"
        # part. So give it 'a' permission to add the part.
        # otherwise, it there will be errors
        ds = self.load_asdf(input_asdf, mode='a')

        # update param based on event information
        update_param(ds.events[0], param)
        # check final param to see if the keys are right
        check_param_keywords(param)

        process_function = \
            partial(process_wrapper, param=param)

        tag_map = {input_tag: output_tag}
        ds.process(process_function, output_asdf, tag_map=tag_map)

        del ds
